import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.*;
import java.util.*;
class Gaussian {

    // return phi(x) = standard Gaussian pdf
    public static double phi(double x) {
        return Math.exp(-x*x / 2) / Math.sqrt(2 * Math.PI);
    }

    // return phi(x) = Gaussian pdf with mean mu and stddev sigma
    public static double phi(double x, double mu, double sigma) {
        return phi((x - mu) / sigma) / sigma;
    }

    // return Phi(z) = standard Gaussian cdf using Taylor approximation
    public static double Phi(double z) {
        if (z < -8.0) return 0.0;
        if (z >  8.0) return 1.0;
        double sum = 0.0, term = z;
        for (int i = 3; sum + term != sum; i += 2) {
            sum  = sum + term;
            term = term * z * z / i;
        }
        return 0.5 + sum * phi(z);
    }



    // return Phi(z, mu, sigma) = Gaussian cdf with mean mu and stddev sigma
    public static double Phi(double z, double mu, double sigma) {
        return Phi((z - mu) / sigma);
    } 

}

enum FileType{SalmonFile15Columns,SalmonFile13Columns,SalmonFileVarColumns,Pictures,SalmonMovingGaussian};
enum RecordType{CommaSeparated,SpaceSeparated};
/**
 * This class handles loading different kind of data sets for input
 * @author osoysal
 *
 */
class EventDB
{
	static Vector<EventDB> currentDBs=new Vector<EventDB>();
	static EventDB createEventDB(String fileName,int nRows,int RowSize,double timeStrechFactor,long seed, double sourceSpeed, double sourceVariance)
	{
		for (int i=0;i<currentDBs.size();i++)
			if (currentDBs.get(i).fileName.equals(fileName))
				return currentDBs.get(i);
		EventDB e=new EventDB(fileName,nRows,RowSize,timeStrechFactor, seed, sourceSpeed, sourceVariance);
		currentDBs.add(e);
		return e;
	}
	
	FileType type;
	Vector<EventRecord> records;
	String fileName;
	int nRows;
	int RowSize;
	long seed;
	double timeStrechFactor;
	double sourceSpeed;
	double sourceVariance;
	private EventDB(String fileName,int nRows,int RowSize,double timeStrechFactor, long seed , double sourceSpeed, double sourceVariance)
	{
		init(nRows,RowSize,fileName,timeStrechFactor,seed, sourceSpeed, sourceVariance );
	}
	/**
	 * initiates the DB. Two main kinds of input is possible
	 * Either text files or a directory containing image files. Images files can be sampled with arbitrary grid dimensions. 
	 * There are three kinds of text files. In fixed size text files,  'nRows' and 'RowSize' are ignored.
	 * #v3 files contain data generated from image files, these files contain grid size separately. 
	 * #v4 files are used to create synthetic data. Following values must be given in separate lines 
	 * tMax : maximum time
	 * vSource : maximum speed of sources movement
	 * variance : variance of data generated by source
	 * seed: seed of the random number generated
	 *
	 * @param nRows
	 * @param RowSize
	 * @param fileName
	 * @param timeStrechFactor
	 */
	
	void init(int nRows,int RowSize,String fileName,double timeStrechFactor,long seed, double sourceSpeed, double sourceVariance)
	{
		this.nRows=nRows;
		this.RowSize=RowSize;
		this.fileName=fileName;
		this.timeStrechFactor=timeStrechFactor;
		this.seed=seed;
		this.sourceSpeed=sourceSpeed;
		this.sourceVariance=sourceVariance;
		records = new Vector<EventRecord>();
		if (isDir(fileName))
		{
			loadImageFiles(fileName,nRows,RowSize);
			saveTextFile(fileName+".txt");
			System.err.println("Saved a copy of records in '"+fileName+".txt'");
		}
		else
			loadTextFile(fileName);
		if (type==FileType.SalmonMovingGaussian)
		{
			saveTextFile(fileName+".txt");
			System.err.println("Saved a copy of records in '"+fileName+".txt'");
		}
		if (timeStrechFactor!=1)
		{
			strechTime();
		}
	}
	
	void strechTime()
	{
		long stime=records.get(0).pctimestamp;
		for (int i=0;i<records.size();i++)
		{
			EventRecord r=records.get(i);
			r.pctimestamp=(long)((r.pctimestamp-stime)*timeStrechFactor+r.pctimestamp);
		}
	}
	
	boolean isDir(String fileName)
	{
		File dir = new File(fileName);
		return dir.isDirectory();
	}
	/**
	 * Creates an array containing all event records
	 * @return
	 */
	EventRecord [] getArray()
	{
		EventRecord [] arr=new EventRecord[records.size()];
		for (int i=0;i<records.size();i++)
			arr[i]=records.get(i);
		return arr;
	}
	/**
	 * This function saves the current event records in this file to a new file
	 * @param fileName name of the output file. This will be deleted if exists
	 */
	public void saveTextFile(String fileName)
	{
		try
		{
			FileWriter fw=new FileWriter(fileName);
			BufferedWriter output = new BufferedWriter(fw);
			output.write("#v3");
			output.newLine();
			output.write("nRows "+nRows+" RowSize "+RowSize);
			output.newLine();
			for (int i=0;i<records.size();i++)
			{
				output.write(records.get(i).toString());
				output.newLine();
			}
			output.close();
		}
		catch (IOException e)
		{
			System.err.println("Couldn't create file:'"+fileName+"'");
			e.printStackTrace();
		}
	}
	
	public void createSyntheticData(int tMax,double vSource,double variance,long seed)
	{
		int defaultTimeDifference=1000;
		Random r;
		if (seed<0)
			r=new Random();
		else
			r=new Random(seed);
		Point2D.Double gaussPos=new Point2D.Double();
		gaussPos.y=r.nextDouble()*nRows;
		gaussPos.x=r.nextDouble()*RowSize;
		Point2D.Double gaussTar=new Point2D.Double();
		gaussTar.y=r.nextDouble()*nRows;
		gaussTar.x=r.nextDouble()*RowSize;
		
		
		for (int t=0;t<tMax;t++)
		{
			for (int i=0;i<nRows;i++)
				for (int j=0;j<RowSize;j++)
				{
					double val=Gaussian.phi(i, gaussPos.y, variance)*Gaussian.phi(j, gaussPos.x, variance);
					int state;
					if (val>0.001)
						state=1;
					else
						state=0;
					records.add(new EventRecord(i*RowSize+j,records.size(),1-state,0,defaultTimeDifference*(t+1)));
				}
			if (gaussPos.distance(gaussTar)<vSource)
			{
				gaussPos.setLocation(gaussTar);
				gaussTar.y=r.nextDouble()*nRows;
				gaussTar.x=r.nextDouble()*RowSize;
			}
			else
			{
				Point2D.Double diff=new Point2D.Double();
				diff.x=gaussTar.x-gaussPos.x;
				diff.y=gaussTar.y-gaussPos.y;
				double len=diff.distance(0, 0);
				diff.x/=len;
				diff.y/=len;
				gaussPos.x=gaussPos.x+diff.x*vSource;
				gaussPos.y=gaussPos.y+diff.y*vSource;
			}
		}
	}
	
	/**
	 * This function loads one of the tree input formats. 
	 * First two formats are comma and space separated 
	 * The last format is also comma separated but it also has an additional header
	 * in the following form
	 * #v3
	 * nRows <nRows> RowSize <RowSize>
	 * or following form
	 * #v4
	 * nRows <nRows> RowSize <RowSize>
	 * tMax <tMax>
	 * vSource <vSource>
	 * variance <variance> 
	 * seed <seed>
	 * @param fileName
	 */
	public void loadTextFile(String fileName) 
	{
		try 
		{
			FileReader fr = new FileReader(fileName);
			BufferedReader input = new BufferedReader(fr);

			String s = input.readLine();
			/// expected format is #v3 \n nRows 10 RowSize 10 
			if (s.startsWith("#v3"))
			{
				s=input.readLine(); // skip this initial line
				type=FileType.SalmonFileVarColumns;
				StringTokenizer tok=new StringTokenizer(s,"=: \n");
				tok.nextToken();
				nRows=Integer.parseInt(tok.nextToken());
				tok.nextToken();
				RowSize=Integer.parseInt(tok.nextToken());
				s=input.readLine(); // go to first data line
			}
			else if (s.startsWith("#v4"))
			{
				s=input.readLine();
				type=FileType.SalmonMovingGaussian;
				int tMax=500;
				while(s instanceof String)
				{
					StringTokenizer tok=new StringTokenizer(s,"=: \n");
					if (tok.countTokens()==2)
					{
						String str=tok.nextToken();
						if (str.startsWith("tMax"))
						{
							tMax=Integer.parseInt(tok.nextToken());
						}
						else if (str.startsWith("vSource"))
						{
							sourceSpeed=Double.parseDouble(tok.nextToken());
						}
						else if (str.startsWith("variance"))
						{
							sourceVariance=Double.parseDouble(tok.nextToken());
						}
						else if (str.startsWith("seed"))
						{
							seed=Long.parseLong(tok.nextToken());
						}
					}
					s=input.readLine();
				}
				createSyntheticData(tMax,sourceSpeed,sourceVariance,seed);
			}
			else
			{
				RecordType r = EventRecord.getEventRecordType(s);
				if (r==RecordType.CommaSeparated)
				{
					type=FileType.SalmonFile15Columns;
					nRows=4;
					RowSize=15;
				}
				else
				{
					type=FileType.SalmonFile13Columns;
					nRows=8;
					RowSize=13;
				}
			}

			while (s instanceof String) 
			{
				// System.out.println(s);
				EventRecord e = EventRecord.Parse(s);
				records.add(e);
				// System.out.println(e);
				// System.out.println("--------");
				s = input.readLine();
			}

		} 
		catch (ArrayIndexOutOfBoundsException e) 
		{
			System.err.println("No Input File\n\nFileIO usage: java FileIO <filename>");
			System.exit(1);
		} 
		catch (IOException e) 
		{
			System.err.println("FileNotFound");
			System.exit(1);
		}
		
		System.err.println("File :'" + fileName + "' is loaded with "
				+ records.size() + " event records...OK");
	}
	/**	 construct the for nRows and nCols
	 divide image into nRows by nCols grid
	 each grid cell should contain some useful stuff
	 */
	public void loadImageFiles(String path,int nRows,int nCols)
	{
		int defaultTimeDifference=5000; // default time difference between two samples in seconds
		int [] channels=new int[600*550*4]; // please no problems here
		int [][]lastStatus=new int[nRows][nCols];
		File dir = new File(path);
		String [] imagenames=dir.list();
		for (int i=0;i<imagenames.length;i++)
		{
			String fname=dir+"/"+imagenames[i];
			System.err.print("Current file'"+fname+"': ");
			BufferedImage b=ImageLoader.loadImage(fname, 1);
			System.err.print("Type="+b.getType()+" ");
			if (b.getTransparency()==BufferedImage.BITMASK)
				System.err.print("BitMask");
			else if (b.getTransparency()==BufferedImage.OPAQUE)
				System.err.print("OPAQUE");
			else if (b.getTransparency()==BufferedImage.TRANSLUCENT)
				System.err.print("TRANSLUCENT");
			// this should be a file of type TYPE_INT_ARGB
			Raster ras=b.getData();
			ras.getPixels(0, 0, ras.getWidth(),ras.getHeight(),channels);
			if (i==0)// no message for the first picture
			{
				for (int r=0;r<nRows;r++)
					for (int c=0;c<nCols;c++)
					{
						int x=(int)(600/2+599.0/(nCols-1)*(c-nCols/2)*0.5);
						int y=(int)(550/2+449.0/(nRows-1)*(r-nRows/2)*0.5);
						if (channels[(y*600+x)*4+0]>0)
							lastStatus[r][c]=1;
						else
							lastStatus[r][c]=0;
					}
			}
			else
			{
				for (int r=0;r<nRows;r++)
				{
					for (int c=0;c<nCols;c++)
					{
						int status;
						int x=(int)(600/2+599.0/(nCols-1)*(c-nCols/2)*0.5);
						int y=(int)(550/2+449.0/(nRows-1)*(r-nRows/2)*0.5);
						if (channels[(y*600+x)*4+0]>0)
							status=1;
						else
							status=0;
/*						int message;
						if (status!=lastStatus[r][c])
							message=0; // send start when status changes
						else
							message=1; // otherwise send end*/
						lastStatus[r][c]=status;
						records.add(new EventRecord(r*nCols+c,records.size(),1-status,0,defaultTimeDifference*i));
					}
				}
			}
			System.err.println(" (0,0)=["+channels[0]+","+channels[1]+","+channels[2]+","+channels[3]+"]");
		}
	}

}

//id,sequence number,type(0=start, 1=end detection),sensor_time (jiffies: 1 sec=56 jiffies),basestation_pctimestamp (milliseconds)
class EventRecord 
{
	
	public int id;

	public int sequenceNumber;

	public int type;

	public long sensor_time;

	public long pctimestamp;
	
	EventRecord()
	{
	}
	
	EventRecord(int id, int sequenceNumber, int type, long sensor_time, long pctimestamp)
	{
		this.id=id;
		this.sequenceNumber=sequenceNumber;
		this.type=type;
		this.sensor_time=sensor_time;
		this.pctimestamp=pctimestamp;
	}

	static EventRecord Parse(String str) {
		EventRecord e = new EventRecord();
		StringTokenizer strTok;
		RecordType type = getEventRecordType(str);
		if (type == RecordType.CommaSeparated)
			strTok = new StringTokenizer(str, ",");
		else
			strTok = new StringTokenizer(str, " ");
		e.id = Integer.parseInt(strTok.nextToken());
		e.sequenceNumber = Integer.parseInt(strTok.nextToken());
		e.type = Integer.parseInt(strTok.nextToken());
		e.sensor_time = Long.parseLong(strTok.nextToken());
		if (type == RecordType.SpaceSeparated)
			strTok.nextToken(); // in this type there is a redundant line
		e.pctimestamp = Long.parseLong(strTok.nextToken());
		return e;
	}

	static RecordType getEventRecordType(String str) 
	{
		if (str.indexOf(",") >= 0)
			return RecordType.CommaSeparated;
		return RecordType.CommaSeparated;
	}

	public String toString() {
		StringBuffer b = new StringBuffer();
		b.append(id);
		b.append(",");
		b.append(sequenceNumber);
		b.append(",");
		b.append(type);
		b.append(",");
		b.append(sensor_time);
		b.append(",");
		b.append(pctimestamp);
		return b.toString();
	}
}