import java.util.*;
/**
 * This interface is used in decision of next node
 * using the available data
 * 
 * @author osoysal
 *
 */
public interface SimController 
{
	/**
	 * Resets the controller
	 *
	 */
	public void reset();
	/**
	 * The next target node given we are in 'cur'
	 * @param cur current node 
	 * @return next node
	 */
	public SimTreeLeaf getNext(SimTreeLeaf cur);
}

class AStarPath implements Comparable<AStarPath>
{
	private Vector<SimTreeLeaf> path;
	private Vector<SimTreeLeaf> nonvisited;
	double cost;
	double heuristic;
	double heuristicMax;
	
	public AStarPath(SimTreeLeaf l)
	{
		path=new Vector<SimTreeLeaf>();
		path.add(l);
		nonvisited=new Vector<SimTreeLeaf>();
		update();
	}
	
	
	public AStarPath(AStarPath par,SimTreeLeaf l)
	{
		path=new Vector<SimTreeLeaf>();
		nonvisited=new Vector<SimTreeLeaf>();
		path.addAll(par.path);
		path.add(l);
		update();
	}
	
	void update()
	{
		nonvisited.clear();
		DistanceMatrix D=SimKernel.theKernel.treeRoots.get(0).D;///WARNING
		boolean visited[]=new boolean[D.nodes.size()];
		for (int i=0;i<D.nodes.size();i++)
			visited[i]=false;
		for (int i=0;i<path.size();i++)
			visited[path.get(i).DistanceMatrixIndex]=true;
		for (int i=0;i<D.nodes.size();i++)
			if (!visited[i])
			{
				SimTreeLeaf n=D.nodes.get(i);
				SimTreeLeaf cur=path.get(0);
				if (cur.getRelativeState(n).bufferSize>0)
					nonvisited.add(n);
			}
		Collections.sort(nonvisited,new java.util.Comparator<SimTreeLeaf>() 
		{
			public int compare(SimTreeLeaf l1, SimTreeLeaf l2) 
			{
				SimTreeLeaf cur=path.get(0);
				if (cur.getRelativeState(l1).bufferSize>cur.getRelativeState(l2).bufferSize)
					return -1;
				return 1;
			}
		});
	}
	
	int size()
	{
		return path.size();
	}
	
	SimTreeLeaf get(int i)
	{
		return path.get(i);
	}
	
	void add(SimTreeLeaf l)
	{
		path.add(l);
		update();
	}
	
	void clear()
	{
		path.clear();
		update();
	}
	
	Vector<SimTreeLeaf> getNonVisitedList()
	{
		return nonvisited;
	}
	
	public int compareTo(AStarPath o)
	{
		if (cost+heuristic<o.cost+o.heuristic)
			return -1;
		return 1;
	}
	
	public String toString()
	{
		StringBuilder sb=new StringBuilder();
		for (int i=0;i<path.size();i++)
		{
			sb.append(path.get(i).DistanceMatrixIndex+" ");
		}
		sb.append("("+cost+"+"+heuristic+"="+(cost+heuristic)+")[hmax="+heuristicMax+"]");
		return sb.toString();
	}
}
/**
 * This abstact class is used in A* algorithm for different cost types. 
 * It defines both the cost of a path (g) and expected cost of the
 * rest of path (h). incremental versions of these functions are also presented where parent 
 * path is presented with the next node.
 * hMax is the upper bound of this one   
 * @author osoysal
 *
 */
abstract class AStarCostFunction
{
	public abstract double g(AStarPath p);
	public abstract double h(AStarPath p);
	public abstract double hMax(AStarPath p);
	public abstract double g(AStarPath p, AStarPath parent);
	public abstract double h(AStarPath p, AStarPath parent);
	public abstract double hMax(AStarPath p, AStarPath parent);
	public void fillValues(AStarPath p)
	{
		p.cost=g(p);
		p.heuristic=h(p);
		p.heuristicMax=hMax(p);
	}
	public void fillValues(AStarPath p,AStarPath parent)
	{
		p.cost=g(p,parent);
		p.heuristic=h(p,parent);
		p.heuristicMax=hMax(p,parent);
	}
}
/**
 * This is the implementation of Simple Minimum Latency heuristic 
 * Data rates and communication is ignored. Also total cost
 * of communication is not considered. 
 * 
 * @author osoysal
 *
 */
class SimpleML extends AStarCostFunction
{
	int distcount[]=new int [40];
	
	public double g(AStarPath p)
	{
		double cost=0;
		if (p.size()==0)
			return 0; 
		SimTreeLeaf cur=p.get(0);
		for (int i=0;i<p.size();i++)
		{
			SimTreeLeaf n=p.get(i);
			cost+=i*cur.getRelativeState(n).bufferSize;
		}
		return cost;
	}
	
	public double h(AStarPath p)
	{
		double cost=0;
		DistanceMatrix D=SimKernel.theKernel.treeRoots.get(0).D; ///////WARNING
		Vector<SimTreeLeaf> nv=p.getNonVisitedList();
		SimTreeLeaf cur=p.get(0); // path really shouldn't be fully empty
									// we need initial node
		SimTreeLeaf pos=p.get(p.size()-1);
		for (int i=0;i<40;i++)
		{
			distcount[i]=0;
		}

		for (int i=0;i<nv.size();i++)
		{
			SimTreeLeaf n=nv.get(i);
			int d=D.getDistance(pos,n);
			if (distcount[d]<=3)
				cost+=(d+distcount[d]*2+p.size())*cur.getRelativeState(n).bufferSize;
			else 
				cost+=(d+6+(distcount[d]-3)*4+p.size())*cur.getRelativeState(n).bufferSize;
			distcount[d]++;
		}
		return cost;
	}
	
	public double g(AStarPath p, AStarPath parent)
	{
		double cost=parent.cost;
		SimTreeLeaf cur=p.get(0);
		for (int i=0;i<parent.size();i++)
			if (parent.get(i)==p.get(p.size()-1))
				return cost;
		cost+=parent.size()*cur.getRelativeState(p.get(p.size()-1)).bufferSize;
		return cost;
	}
	
	public double h(AStarPath p, AStarPath parent)
	{
		// too complicated due to changes in relative position and distance
		return h(p);
	}
	
	public double hMax(AStarPath p)
	{
		double cost=0;
		int dist=0;
		DistanceMatrix D=SimKernel.theKernel.treeRoots.get(0).D; /////////WARNING
		Vector<SimTreeLeaf> nv=p.getNonVisitedList();
		SimTreeLeaf cur=p.get(0); // path really shouldn't be fully empty
									// we need initial node
		SimTreeLeaf pos=p.get(p.size()-1);
		for (int i=0;i<nv.size();i++)
		{
			SimTreeLeaf n=nv.get(i);
			dist+=D.getDistance(pos,n); // go to node
			cost+=(dist+p.size())*cur.getRelativeState(n).bufferSize;
			pos=n;
		}
		return cost;
	}
	
	public double hMax(AStarPath p, AStarPath parent)
	{
		return hMax(p);
	}
	
}

/**
 * This class implements A star algorithm with different heuristics.
 *  
 * @author osoysal
 *
 */
class AStarController implements SimController   
{
	SimKernel krn;
	AStarCostFunction func;
	AStarPath p;
	int counter;
	SimTreeLeaf lastRoot;
	
	public AStarController()
	{
		krn=SimKernel.theKernel;
		func=new SimpleML(); // this should be more intelligent
		lastRoot=null;
		counter=0;
	}
	
	AStarPath search(SimTreeLeaf start)
	{
		PriorityQueue<AStarPath> pq=new PriorityQueue<AStarPath>();
		AStarPath cur=new AStarPath(start);
		AStarPath best=null;
		AStarPath p;
		func.fillValues(cur);
		double bound=cur.cost+cur.heuristicMax;
		pq.add(cur);
		int nodesExpanded=0;
		do
		{
			cur=pq.remove();
			nodesExpanded++;
			if (nodesExpanded%100==99)
			{
				Iterator<AStarPath> i;
				AStarPath a;
				for (i=pq.iterator();i.hasNext();)
				{
					a=i.next();
					if (a.cost+a.heuristic>bound)
						pq.remove(a);
				}
			}
			if (cur.heuristic==0)
				break; // this means we have found the result
			if (cur.cost+cur.heuristic>bound)
				continue;
			SimTreeLeaf curNode=cur.get(cur.size()-1);
			if (cur.size()<12)
			{
				if (curNode.parent!=null )
				{
					p=new AStarPath(cur,curNode.parent);
					func.fillValues(p,cur);
					if (bound>p.cost+p.heuristic)
					{
						if (p.cost+p.heuristicMax<bound)
							bound=p.cost+p.heuristicMax;
						pq.add(p);
					}
				}
				for (int i=0;i<curNode.children.length;i++)
				{
					p=new AStarPath(cur,curNode.children[i]);
					func.fillValues(p,cur);
					if (bound>p.cost+p.heuristic)
					{
						if (p.cost+p.heuristicMax<bound)
							bound=p.cost+p.heuristicMax;
						pq.add(p);
					}
				}
			}
			else
			{
				if (best == null || best.cost+best.heuristic*100>cur.cost+cur.heuristic*100)
					best=cur;
			}
		}while (!pq.isEmpty());
		System.out.println("Nodes Expanded: "+nodesExpanded+"      isQueue Empty? "+pq.isEmpty());
		if (cur.heuristic == 0 || best==null || best.cost+best.heuristic*100>cur.cost+cur.heuristic*100)
			best=cur;
		return best;
	}
		
	public void reset()
	{
		// nothing it is stateless
	}

	public SimTreeLeaf getNext(SimTreeLeaf cur)
	{
		if (p==null)
		{
			p=search(cur);
			counter=0;
		}
		System.out.println(p);
		if (p==null)
			return cur;
		else
		{
			if (p.get(counter)==cur)
				counter++;
			if (counter>=p.size())
			{
				p=null;
				return cur;
			}
			else
				return p.get(counter);
		}
	}
	
}

/**
 * This is the idealized Salmon Controller. 
 * The controller can instantly know the data rates
 * a more realistic version would only know the relative data rates
 * I'm still depending on the old implementation for this part
 * a more elegant way to do this would be through state histories 
 * stored in the nodes. 
 * 
 * @author osoysal
 *
 */
class IdealSalmonController implements SimController
{
	public void reset()
	{
		// nothing it is stateless
	}

	public SimTreeLeaf getNext(SimTreeLeaf cur)
	{
		SimTreeLeaf targetNode=cur;
		double eps = cur.tree.root.getDownStreamDataRate();////WARNING
		if (cur.getUpStreamDataRate() > eps / 2)
			targetNode = cur.parent;
		for (int i = 0; i < cur.children.length; i++) 
		{
			if (cur.children[i].getDownStreamDataRate() > eps / 2)
				targetNode = cur.children[i];
		}
		if (SimKernel.theKernel.currentEvent>=SimKernel.theKernel.eventList.length)
			return (new MaxBufferSizeController()).getNext(cur);
		else
			return targetNode;
	}
}

/**
 * Total buffer size heuristic.
 * 
 * This heuristic moves the MB to the center of mass of the buffers.
 * I'm still depending on the old implementation for this part
 * a more elegant way to do this would be through state histories 
 * stored in the nodes. 
 * @author osoysal
 *
 */
class TotalBufferSizeController implements SimController
{
	public void reset()
	{
		// nothing it is stateless
	}

	public SimTreeLeaf getNext(SimTreeLeaf currentNode)
	{
		SimTreeLeaf targetNode=currentNode;
		int nneigh=currentNode.children.length;
		if (currentNode.parent!=null)
			nneigh++;
		SimTreeLeaf neigh[]=new SimTreeLeaf[nneigh];
		for (int i=0;i<currentNode.children.length;i++)
			neigh[i]=currentNode.children[i];
		if (currentNode.parent!=null)
			neigh[nneigh-1]=currentNode.parent;
		double childbuffers[]=new double[nneigh];
		double eps=0;
		for (int i=0;i<nneigh;i++)
			childbuffers[i]=0;
		for (int i=0;i<currentNode.tree.nodeList.size();i++)
		{
			SimTreeLeaf n=currentNode.tree.nodeList.get(i);
			int buf=currentNode.getRelativeState(currentNode.tree.nodeList.get(i)).bufferSize;
			eps+=buf;
			if (currentNode.childlevel(n)>=0)
			{
				for (int j=0;j<currentNode.children.length;j++)
				{
					if (currentNode.children[j].childlevel(n)>=0)
						childbuffers[j]+=buf;
				}
			}
			else
			{
				childbuffers[nneigh-1]+=buf;
			}
		}
		for (int i = 0; i < nneigh; i++) 
		{
			if (childbuffers[i] > eps / 2)
				targetNode = neigh[i];
		}
		return targetNode;
	}
}

/**
 * Maximum buffer size heuristic.
 * 
 * This heuristic moves the MB toward the maximum perceived buffer.
 * @author osoysal
 *
 */
class MaxBufferSizeController implements SimController
{
	public void reset()
	{
		// nothing it is stateless
	}

	public SimTreeLeaf getNext(SimTreeLeaf currentNode)
	{
		SimTreeLeaf MaxNode=currentNode;
		SimTree t=currentNode.tree;
		for (int i=0;i<t.nodeList.size();i++)
		{
			if (currentNode.getRelativeState(t.nodeList.get(i)).bufferSize>MaxNode.bufferSize)
				MaxNode=t.nodeList.get(i);
		}
		
		SimTreeLeaf targetNode=currentNode;
		boolean isChild=false;
		for (int i=0;i<currentNode.children.length;i++)
			if (currentNode.children[i].childlevel(MaxNode)>=0) // it is our child
			{
				targetNode=currentNode.children[i];
				isChild=true;
				break;
			}
		if (!isChild) // it is our parent
			targetNode=currentNode.parent;
		if (targetNode==null)
			targetNode=currentNode;
		return targetNode;
	}
}

/**
 * Weighted buffer size heuristic.
 *
 * This heuristic weights the buffer sizes with their distance.
 * So closer clusters are prefered over larger but further clusters. 
 * I'm still depending on the old implementation for this part
 * a more elegant way to do this would be through state histories 
 * stored in the nodes. 
 * @author osoysal
 *
 */
class WeightedBufferSizeController implements SimController
{
	public void reset()
	{
		// nothing it is stateless
	}

	public SimTreeLeaf getNext(SimTreeLeaf currentNode)
	{
		SimTreeLeaf targetNode=currentNode;
		// could be a lot simpler if i used
		// a neighbour array filled with parent/children
		// and it would be crystal clear. 
		int r=currentNode.tree.root.depth();
		int n=currentNode.children.length;
		SimTree tree=currentNode.tree;
		double max[][]=new double[n+1][2*r+2]; // max[i][j]: moving 
											// j depth passing from i'th neighbor
											// 0th column is useless
		for (int i=0;i<n+1;i++)
			for (int j=0;j<2*r+2;j++)
				max[i][j]=0;
		
		for (int i=0;i<tree.nodeList.size();i++)
		{
			int d=tree.D.getDistance(currentNode, tree.nodeList.get(i));
			if (d>0)
			{
				if (currentNode.parent!=null && currentNode.parent.isOnPath(currentNode, tree.nodeList.get(i)))
				{
					if (max[n][d]<currentNode.getRelativeState(tree.nodeList.get(i)).bufferSize)
						max[n][d]=currentNode.getRelativeState(tree.nodeList.get(i)).bufferSize;
				}
				for (int j=0;j<n;j++)
				{
					if (currentNode.children[j].isOnPath(currentNode, tree.nodeList.get(i)))
					{
						if (max[j][d]<currentNode.getRelativeState(tree.nodeList.get(i)).bufferSize)
							max[j][d]=currentNode.getRelativeState(tree.nodeList.get(i)).bufferSize;
					}
				}
			}
		}
		
		
		double fit[]=new double[n+1];
		if (currentNode.parent!=null)
		{
			fit[n]=0;
			for (int i=1;i<2*r+2;i++)
				fit[n]+=max[n][i]*1.0/(i*i);
		}
		for (int i=0;i<currentNode.children.length;i++)
		{
			fit[i]=0;
			for (int j=1;j<2*r+2;j++)
				fit[i]+=max[i][j]*1.0/(j*j);
		}
		int fitidx=0;
		for (int i=0;i<n+1;i++)
			if (fit[i]>fit[fitidx])
				fitidx=i;
		if (fitidx==n)
			targetNode=currentNode.parent;
		else
			targetNode=currentNode.children[fitidx];
		return targetNode;
	}
}

/**
 * Random Walk heuristic.
 *
 * This heuristic moves the robot on the tree randomly. 
 * Parent has double the chance of being chosen as the next node than any child
 * this helps to keep the BS away from the leaves. 
 *  
 * @author osoysal
 *
 */
class RandomWalkController implements SimController
{
	public void reset()
	{
		// nothing it is stateless
	}

	public SimTreeLeaf getNext(SimTreeLeaf currentNode)
	{
		SimTreeLeaf targetNode=currentNode;
		int range=currentNode.children.length;
		if (currentNode.parent!=null)
			range+=2;
		int r=SimKernel.theKernel.rnd.nextInt(range);
		if (r>=currentNode.children.length)
			targetNode=currentNode.parent;
		else
			targetNode=currentNode.children[r];
		return targetNode;
	}
}
/**
 * This is the simplest fisher controller which moves along all cluster heads one by one 
 * @author osoysal
 *
 */
class SimpleFisherController implements SimController
{
	enum State{GoToTree,GoToRp,RunTreeWalk};
	State curState;
	
	int lastTargetIdx; // last tree i chose as target
	SimTreeLeaf lastKnownArrowRoots[];
	long lastVisitTime[];
	SimKernel krn;
	SimTreeLeaf lastTarget;
	WeightedBufferSizeController treeWalker;
	long treeTraversalTimer; // this is the duration passed for traversing the tree
	long lastControlStepTime; // this is the last time control step is called.
	long lastStepDuration; // duration of the last step 
	
	SimpleFisherController()
	{
		krn=SimKernel.theKernel;
		treeWalker=new WeightedBufferSizeController();
		lastKnownArrowRoots=new SimTreeLeaf[krn.treeRoots.size()];
		lastVisitTime=new long[krn.treeRoots.size()];
		reset();
	}
	
	public int getNextIdx()
	{
		return (lastTargetIdx+1)%krn.treeRoots.size();
	}
	public void reset()
	{
		for (int i=0;i<krn.treeRoots.size();i++)
		{
			lastKnownArrowRoots[i]=krn.treeRoots.get(i).root;
			lastVisitTime[i]=-1;
		}
		lastTargetIdx=0;
		lastTarget=lastKnownArrowRoots[lastTargetIdx];
		treeTraversalTimer=-1;
		lastControlStepTime=-1;
		lastStepDuration=0;
		curState=State.GoToTree;
	}
	
	public SimTreeLeaf getNext(SimTreeLeaf currentNode)
	{
		int idx=krn.treeRoots.indexOf(currentNode.tree);
		lastVisitTime[idx]=krn.currentTime;
		if (lastControlStepTime>0)
			lastStepDuration=krn.currentTime-lastControlStepTime;
		switch(curState)
		{
		case GoToTree:
			if (idx==lastTargetIdx)
			{
				curState=State.GoToRp;
				lastKnownArrowRoots[lastTargetIdx]=currentNode.tree.arrowRoot;
				lastTarget=lastKnownArrowRoots[lastTargetIdx];
			}
			break;
		case GoToRp:
			if (lastTarget==currentNode && currentNode.bufferSize<SimTreeLeaf.LinkCapacity*2)
			{
				curState=State.RunTreeWalk;
				treeTraversalTimer=0;
			}
			break;
		case RunTreeWalk:
			treeTraversalTimer+=lastStepDuration;
			lastTarget=treeWalker.getNext(currentNode);
			if (treeTraversalTimer>=Salmon.treeTourLength/krn.treeRoots.size())
			{
				curState=State.GoToTree;
				treeTraversalTimer=-1;
				lastTargetIdx=getNextIdx();
				lastTarget=lastKnownArrowRoots[lastTargetIdx];				
			}
		}
		lastControlStepTime=krn.currentTime;
		return lastTarget;
	}
}
/**
 * A bit more intelligent fisher controller which first solves 
 * TSP for the tree roots and uses this order for deciding on the visit order of trees
 * @author osoysal
 *
 */
class TSPFisherController extends SimpleFisherController
{
	double dist[][]; // distances between tree roots
	int visitorder [] ;
	TSPFisherController()
	{
		super();
		int n=lastKnownArrowRoots.length;
		visitorder=new int [n];
		dist=new double[n][n];
		calculateRoute();
	}
	
	void calculateRoute()
	{
		int n=lastKnownArrowRoots.length;
		for (int i=0;i<n;i++)
			for (int j=0;j<n;j++)
			{
				dist[i][j]=lastKnownArrowRoots[i].pos.distance(lastKnownArrowRoots[j].pos);
			}
		solveTSP();
//		System.out.println("Current Time: "+SimKernel.theKernel.currentTime/1000.0);
	}
	
	double cost(int order[])
	{
		double c=0;
		for (int i=1;i<order.length;i++)
			c+=dist[order[i]][order[i-1]];
		c+=dist[order[0]][order[order.length-1]];
		return c;
	}
	
	void getNext(int indices[])
	{
        // find the rightmost element that is smaller than the element at its right
        int i = indices.length - 1;
        while (indices[i - 1] >= indices[i])
                i = i - 1;
        // find the rightmost element that is bigger then the other one
        int j = indices.length;
        while (indices[j - 1] <= indices[i - 1])
                j = j - 1;
        // swap them (always is i < j)
        //swap(i - 1, j - 1);
        int tmp=indices[i-1];
        indices[i-1]=indices[j-1];
        indices[j-1]=tmp;
        // now the elements at the right of i
        // are in descending order, so  reverse them all
        i++;
        j = indices.length;
        while (i < j) 
        {
//                swap(i - 1, j - 1);
            tmp=indices[i-1];
            indices[i-1]=indices[j-1];
            indices[j-1]=tmp;
                i++;
                j--;
        }		
	}
	void solveTSP()
	{
		double best;
		double current;
		int n=lastKnownArrowRoots.length;
		int order[]=new int[n];
		for (int i=0;i<visitorder.length;i++)
		{
			visitorder[i]=i;
			order[i]=i;
		}
		best=cost(visitorder);
		int perms=0;
		while (order[0]==0)
		{
			getNext(order);
			perms++;
			current=cost(order);
			if (current<best)
			{
				best=current;
				for (int i=0;i<order.length;i++)
					visitorder[i]=order[i];
			}
		}
		System.err.println("nperms:"+perms);
		for (int i=0;i<n;i++)
		{
			System.err.print((char)('A'+visitorder[i]));
		}
		System.err.println();
	}
	
	public int getNextIdx() // override parent method for choosing a better route
	{
		int idx;
		for (idx=0;lastTargetIdx!=visitorder[idx];idx++)
		{}
		if (idx==visitorder.length-1)
		{
			idx=0;
			//calculateRoute();
//			System.out.println("Current Time: "+SimKernel.theKernel.currentTime/1000.0);
		}
		else
			idx++;
		return visitorder[idx];
	}
}
/**
 * This is more intelligent that takes into account the buffer sizes
 * so the cost is not a simple addition of distances but the expected package latency 
 * this still doesn't consider the possible cycles in the route. 
 * cycles are too complicated to attack right now. 
 * @author osoysal
 *
 */
class MLPFisherController extends TSPFisherController
{
	int lastCollectedPackages[];
	int lastPackageCounter[];
	long lastDuration[];
	MLPFisherController()
	{
		super();
		int n=lastKnownArrowRoots.length;
		for (int i=0;i<n;i++)
			for (int j=0;j<n;j++)
			{
				dist[i][j]=lastKnownArrowRoots[i].pos.distance(lastKnownArrowRoots[j].pos);
			}
	}

	void calculateRoute()
	{
		if (lastCollectedPackages==null)
		{
			lastCollectedPackages=new int[lastKnownArrowRoots.length];
			lastPackageCounter=new int[lastKnownArrowRoots.length];
			lastDuration= new long[lastKnownArrowRoots.length];
			for (int i=0;i<lastKnownArrowRoots.length;i++)
			{
				lastCollectedPackages[i]=1;
				lastPackageCounter[i]=-1;
				lastVisitTime[i]=-1;
				lastDuration[i]=1;
			}			
		}
		if (lastVisitTime[0]>=0)
		{
			Statistics s=SimKernel.theKernel.stats;
			Vector<SimTree> trees=SimKernel.theKernel.treeRoots;
			for (int i=0;i<lastKnownArrowRoots.length;i++)
			{
				lastCollectedPackages[i]=0;
			}
			// check package history for delivered packages
			for (int i=0;i<s.allPackages.size();i++)
			{
				if (s.allPackages.get(i).deliveryTime>=lastVisitTime[0])
				{
					for (int j=0;j<trees.size();j++)
						if (s.allPackages.get(i).initiator.tree==trees.get(j))
						{
							lastCollectedPackages[j]++;
							break;
						}
				}
			}
			for (int i=1;i<lastKnownArrowRoots.length;i++)
			{
				lastDuration[visitorder[i]]=lastVisitTime[visitorder[i]]-lastVisitTime[0];
			}
			lastDuration[0]=1; // this is not relevant
		}
		int n=lastKnownArrowRoots.length;
		for (int i=0;i<n;i++)
			for (int j=0;j<n;j++)
			{
				dist[i][j]=lastKnownArrowRoots[i].pos.distance(lastKnownArrowRoots[j].pos);
			}
		solveTSP();
//		System.out.println("Current Time: "+SimKernel.theKernel.currentTime/1000.0);

//		super.calculateRoute();
	}
	
	double cost(int order[]) // override cost function from the parent
	{
		double c=0;
		double latency=0;
		
		for (int i=1;i<order.length;i++)
		{
			latency+=(dist[order[i]][order[i-1]]);
			c+=latency*(lastCollectedPackages[order[i]]);///lastDuration[order[i]]);
		}
		return c;
	}

	public int getNextIdx() // override parent method for choosing a better route
	{
		Salmon fish=SimKernel.theKernel.theFish;
		lastPackageCounter[lastTargetIdx]=fish.collectedPackages;
		int idx;
		for (idx=0;lastTargetIdx!=visitorder[idx];idx++)
		{}

		if (idx==visitorder.length-1)
		{
			idx=0;
			calculateRoute();
		}
		else
			idx++;
		return visitorder[idx];
	}
}

