import java.util.*;
import java.awt.geom.*;

class StateHistoryElement
{
	int bufferSize; // the size of buffer at current time step
	int dataRate; // the rate of data at current time step
	boolean isReal; /// whether this element corresponds to a real data element or a stub 
	public StateHistoryElement(int bufferSize, int dataRate)
	{
		this.bufferSize=bufferSize;
		this.dataRate=dataRate;
		isReal=true;
	}
	
	public StateHistoryElement()
	{
		this.bufferSize=0;
		this.dataRate=0;
		isReal=false;		
	}
}
/**
 * This class represents a tree for salmon algorithm.
 * Multiple trees are also supported. One important note
 * about this class is that this class must be constructed in three steps
 * 
 * 1) allocation: This handles the basic allocations
 * 2) tree construction: Here the tree connections will be created using 'growTree'
 * 3) update: This step will update the distance matrix representation for the tree 
 * 
 * This rather complicated construction is necessary to provide support for multiple trees.
 * By having multiple tree construction steps, we can grow multiple trees similar to the 
 * natural way of flooding.  
 * @author osoysal
 *
 */
class SimTree
{
	//public int nodeState[][];
	
	Vector<SimTreeLeaf> nodeList; //the list of nodes in this tree for linear access 
	
	DistanceMatrix D;
	
	SimTreeLeaf root;
	
	SimTreeLeaf arrowRoot;
	
	
	public SimTree(SimTreeLeaf root)
	{
		setRoot(root);
		root.tree=this;
		nodeList=new Vector<SimTreeLeaf>();
		nodeList.add(root);
		D = new DistanceMatrix();
	}
	void setRoot(SimTreeLeaf root)
	{
		this.root=root;
		this.arrowRoot=root;
	}
	
	void reset()
	{
		for (int i=0;i<nodeList.size();i++)
			nodeList.get(i).reset();
	}
	
	void update()
	{
		reset();
		D.fillList(root);
		D.calculateDistances();		
	}
	
	void growTree()
	{
		Random rnd=SimKernel.theKernel.rnd;
		SimKernel krn=SimKernel.theKernel;
		int idx = (int) (rnd.nextDouble() * rnd.nextDouble() * nodeList.size());
		SimTreeLeaf l = (SimTreeLeaf) nodeList.get(idx);
		Vector<SimTreeLeaf> neighs=krn.getNodesInRange(l);
		for (int i=0;i<neighs.size();i++)
			if (neighs.get(i).isIsolated())
			{
				// add as a child
				SimTreeLeaf child=neighs.get(i);
				child.tree=this;
				l.addChild(child);
				nodeList.add(child);
			}
	}
	
	boolean contains(SimTreeLeaf n)
	{
		for (int i=0;i<nodeList.size();i++)
			if (nodeList.get(i)==n)
				return true;
		return false;
	}
}

class SimTreeLeaf // tree node for the embedding
{
	static final SimTreeLeaf simNull = new SimTreeLeaf();

	static int currentNodeId = 0;
	
	static int MaxBufferSize = 100; // all messages after this amount is discarded
	
	static int LinkCapacity = 8; // each link can forward at most this amount of messages in unit time
	
	static boolean superFastBS = false; // if true base station has 10x link capacity

	static int bufferThreshold=0; /// the threshold after node starts sending messages
	
	static int ThresholdRotSpeed=1; /// the speed for threshold reduction
	
	static boolean trackrobot = false; // whether the arrow protocol follows the robot

	int nodeId;
	
	int DistanceMatrixIndex;

	MessageQueue pkgQueue; /// queue of the packages
	
	LinkedList<StateHistoryElement> stateHistory;

	SimTreeLeaf[] children;

	SimTreeLeaf parent;

	SimTreeLeaf arrowNext; // next node in arrow protocol
	
	SimTree tree;

	Point2D.Double pos;
	
	int state; // state of the node
	
	int thresholdCounter; // in order to make all data sent eventually
	
	int bufferSize; // current size of buffer
	
	int recvQueue; // messages to be received next time step
	
//	int downStreamBufferSize; // the total buffer size of downstream nodes including this one

	double usage;
	
	int level;

	public SimTreeLeaf(SimTree tree,double x, double y) 
	{
		init(tree,x,y);
	}

	public SimTreeLeaf() 
	{
		init(null,0,0);
	}

	void init(SimTree tree,double x, double y)
	{
		this.tree=tree;
		pos=new Point2D.Double(x,y);
		children = new SimTreeLeaf[0];
		parent = null;
		arrowNext = null;
		level = 0;
		nodeId = currentNodeId++; //WARNING... this can be a trouble maker
		pkgQueue=new MessageQueue();
		stateHistory=new LinkedList<StateHistoryElement>();
		reset();
	//	bufferThreshold=400; // forward all you have
	}
	
	void reset() // don't change topology but reset everything else
	{
		state=0;
		bufferSize=0;
		recvQueue=0;
		usage=0;
		thresholdCounter=0;
		arrowNext=parent;
		pkgQueue.reset();
		stateHistory.clear();
	}

	void addChild(SimTreeLeaf n) {
		int currentchild = 0;
		if (children == null) {
			children = new SimTreeLeaf[1];
		} else {
			SimTreeLeaf[] tmp = new SimTreeLeaf[children.length + 1];
			for (int i = 0; i < children.length; i++)
				tmp[i] = children[i];
			currentchild = children.length;
			children = tmp;
		}
		children[currentchild] = n;
		n.setParent(this);
	}
	
	void setParent(SimTreeLeaf parent)
	{
		this.parent=parent;
		arrowNext=parent;
		if (parent!=null)
			level=parent.level+1;
		else
			level=0;
	}

	void removeChild(SimTreeLeaf n) 
	{
		SimTreeLeaf[] tmp = new SimTreeLeaf[children.length - 1];
		for (int i = 0, j = 0; i < children.length; i++)
			if (children[i] != n)
				tmp[j++] = children[i];
		children = tmp;
	}

	public String toString() 
	{
		String str = "";
		str = str + nodeId + ": ";
		return str;
	}

	int childlevel(SimTreeLeaf descendant) 
	{
		if (this == descendant)
			return 0;
		if (children.length == 0)
			return -1; // not a children
		for (int i = 0; i < children.length; i++) 
		{
			int res = children[i].childlevel(descendant);
			if (res >= 0)
				return res + 1;
		}
		return -1; // not a grand children
	}
	
	boolean isIsolated()
	{
		for (int i=0;i<SimKernel.theKernel.treeRoots.size();i++)
			if (this==SimKernel.theKernel.treeRoots.get(i).root)
				return false;
		return (parent==null) && (children.length==0);
	}
	
	
	SimTreeLeaf lowestCommonAncestor(SimTreeLeaf l)
	{
		if (childlevel(l)>=0)
			return this;
		else if (l.childlevel(this)>=0)
			return l;
		return parent.lowestCommonAncestor(l);
	}
	
	boolean isOnPath(SimTreeLeaf s,SimTreeLeaf e)
	{
		SimTreeLeaf c=e.lowestCommonAncestor(s);
		if (c.childlevel(this)>=0)
		{
			if (this.childlevel(s)>=0 || this.childlevel(e)>=0)
				return true;
			else
				return false;
		}
		else
			return false;
	}
	
	public double getDownStreamDataRate() // data coming from the children
	{
		double sum = state;
		for (int i = 0; i < children.length; i++) {
			sum += children[i].getDownStreamDataRate();
		}
		return sum;
	}

	public double getUpStreamDataRate() // data coming from parent
	{
		return tree.root.getDownStreamDataRate() - getDownStreamDataRate();
	}

	// here we need a path based cost function
	// how can be obtain that?  down stream data and some stuff can be useful
	// here we wonder the usage of node n when basestation is in station
	public double getNodeUsage(SimTreeLeaf station)
	{
		SimTreeLeaf treeRoot=tree.root;
		//int idx=D.getIdx(station);
		int childl=childlevel(station);
		if (childl>0) // n is an ancestor of basestation
		{ // actually this is the most difficult case probably
			int stationParentIdx=0;
			for (int i=0;i<children.length;i++)
				if (children[i].childlevel(station)>=0)
				{
					stationParentIdx=i;
					break;
				}
			return (treeRoot.getDownStreamDataRate()-children[stationParentIdx].getDownStreamDataRate() -state)*2+state; // only this child is not forwarding to this node. 
		}
		else if (childl==0) // i'm the base station
		{
			return (treeRoot.getDownStreamDataRate()-state)*2+state; // total data rate
		}
		else // n is a child of base station or we share an ancestor 
		{
			return (getDownStreamDataRate()-state)*2+state; // i only forward my downstream data
		}
	}

	public boolean isOptimal() 
	{
		SimTreeLeaf treeRoot=tree.root;
		double eps = treeRoot.getDownStreamDataRate();
		if (getUpStreamDataRate() > eps / 2)
			return false;
		for (int i = 0; i < children.length; i++) {
			if (children[i].getDownStreamDataRate() > eps / 2)
				return false;
		}
		return true;
	}
	
	public SimTreeLeaf getOptimalNode() // based on this tree node
	// just find one in
	// rightmost
	{
		if (isOptimal())
			return this;
		for (int i = 0; i < children.length; i++) {
			SimTreeLeaf l = children[i].getOptimalNode();
			if (l != null)
				return l;
		}
		return null;
	}
	
	/**
	 * This function should only be called for the arrowRoot
	 *
	 */
	void updateArrowDataRate()
	{
		// update arrows
		SimTreeLeaf targetNode=null;
		double eps = tree.root.getDownStreamDataRate();
		if (getUpStreamDataRate() > eps / 2)
			targetNode = parent;
		for (int i = 0; i < children.length; i++) 
		{
			if (children[i].getDownStreamDataRate() > eps / 2)
				targetNode = children[i];
		}
		if (targetNode!=null)
		{
			targetNode.arrowNext=null;
			this.arrowNext=targetNode;
			tree.arrowRoot=targetNode;
		}
	}
	
	void updateArrowBufferSize()
	{
		SimTreeLeaf targetNode=null;
		double eps = 0;
		double childbuffers[]=new double[children.length];
		double totalchild=0;
		for (int i=0;i<children.length;i++)
			childbuffers[i]=0;
		for (int i=0;i<tree.nodeList.size();i++)
		{
			int buf=getRelativeState(tree.nodeList.get(i)).bufferSize;
			eps+=buf;
			for (int j=0;j<children.length;j++)
			{
				if (children[j].childlevel(tree.nodeList.get(i))>=0)
					childbuffers[j]+=buf;
			}
		}
		for (int i=0;i<children.length;i++)
			totalchild+=childbuffers[i];
		
		
		if (eps-(totalchild+bufferSize) > eps / 2)
			targetNode = parent;
		for (int i = 0; i < children.length; i++) 
		{
			if (childbuffers[i] > eps / 2)
				targetNode = children[i];
		}
		if (targetNode!=null)
		{
			targetNode.arrowNext=null;
			this.arrowNext=targetNode;
			tree.arrowRoot=targetNode;
		}
	}
	
	void updateArrowTrackRobot()
	{
		// update arrows
		SimTreeLeaf fishPos=SimKernel.theKernel.theFish.currentNode;
		SimTreeLeaf targetNode=null;
		if (tree.arrowRoot==fishPos)
			return; // nothing to change then
		for (int i = 0; i < children.length; i++) 
		{
			if (children[i].childlevel(fishPos) >=0)
				targetNode = children[i];
		}
		/// this means, no children has the robot
		/// and we have the robot in tree
		if (targetNode==null && tree.root.childlevel(fishPos)>=0)
			targetNode = parent;
		if (targetNode!=null)
		{
			targetNode.arrowNext=null;
			this.arrowNext=targetNode;
			tree.arrowRoot=targetNode;
		}
		
	}
	/** we need two functions to collect sent messages
	 * This structure is present to guarantee the 
	 * delivery of messages without violating the LinkCapacity requirement
	 *
	 */
	public void step() 
	{
		// find where to route the messages
		SimKernel krn=SimKernel.theKernel;
		//SimTreeLeaf basestation=krn.theFish.currentNode;

//		int childl=this.childlevel(bs);
		if (this==tree.arrowRoot)
		{
			if (trackrobot)
				updateArrowTrackRobot();
			else
				updateArrowBufferSize();
		}
		if (pos.distance(krn.theFish.pos)<krn.CommRange/2) // i'm the base station
		{
			int capacity=LinkCapacity;
			double energyMult=1;
			if (superFastBS)
			{
				capacity=LinkCapacity*10;
//				energyMult=0.1;
			}
			if (bufferSize<capacity)
			{
				pkgQueue.sendToBS(bufferSize);
				usage+=(bufferSize*energyMult);
				krn.theFish.collectedPackages+=bufferSize;
				bufferSize=0;
			}
			else
			{
				pkgQueue.sendToBS(capacity);
				usage+=(capacity*energyMult);
				krn.theFish.collectedPackages+=capacity;
				bufferSize-=capacity;
			}
		}
		if (arrowNext!=null)
		{
			if (bufferSize>0)
			{
				if (bufferSize>=bufferThreshold-thresholdCounter/ThresholdRotSpeed)
				{
					if (arrowNext.recvQueue<LinkCapacity)
					{
						if (bufferSize<LinkCapacity-arrowNext.recvQueue)
						{
							usage+=bufferSize; // send whole buffer
							arrowNext.recvQueue+=bufferSize;
							pkgQueue.transferMessages(arrowNext.pkgQueue, bufferSize);
							bufferSize=0;
						}
						else
						{
							usage+=LinkCapacity-arrowNext.recvQueue; // send a part of it 
							bufferSize-=LinkCapacity-arrowNext.recvQueue;
							pkgQueue.transferMessages(arrowNext.pkgQueue, LinkCapacity-arrowNext.recvQueue);
							arrowNext.recvQueue=LinkCapacity;
						}
					}
					thresholdCounter=0;
				}
				else 
				{
					if (arrowNext.recvQueue==0)
						thresholdCounter++;
					else
						thresholdCounter=0;
					if (thresholdCounter*ThresholdRotSpeed>bufferThreshold)
						thresholdCounter=bufferThreshold*ThresholdRotSpeed;
				}
			}
		}
	}
	
	int depth()
	{
		int depth=0;
		for (int i=0;i<children.length;i++)
		{
			int tdepth=children[i].depth();
			if (tdepth>depth)
				depth=tdepth;
		}
		return depth+1;
	}
	
	void fixLevel()
	{
		if (parent!=null)
			level=parent.level+1;
		else
			level=0;
		for (int i=0;i<children.length;i++)
		{
			children[i].setParent(this);
			children[i].fixLevel();
		}
	}
	
	void updateBufferSizeVars()
	{
		
		// here i update the history 
		// history is a simpler way to do this since i don't need to bother with
		// matching nodes and stuff which wouldn't add anything to the simulation
		StateHistoryElement se;
		if (stateHistory.size()==0)
			se=new StateHistoryElement(bufferSize,bufferSize);
		else
			se=new StateHistoryElement(bufferSize,bufferSize-stateHistory.getFirst().bufferSize);
		stateHistory.addFirst(se);
		if (stateHistory.size()>SimKernel.theKernel.nRows*SimKernel.theKernel.RowSize)
			stateHistory.removeLast();
	}
	
	/**
	 * This function checks the state information for the given history.
	 * @param relativeTime relative time in seconds. 0 corresponds to the current
	 * @return StateHistoryElement at given time if present, else 0,0
	 */
	public StateHistoryElement getPastState(int relativeTime)
	{
		final StateHistoryElement snull=new StateHistoryElement();
		if ((-relativeTime)>=stateHistory.size())
			return snull;
		else
			return stateHistory.get(-relativeTime);
	}

	/** 
	 * This function calculates the relative state of node 'n' to this node.
	 * @param n The node that is addressed
	 * @return currently known state of the node 'n' at this node.
	 */
	public StateHistoryElement getRelativeState(SimTreeLeaf n)
	{
		return n.getPastState(-tree.D.getDistance(this, n));
	}
	
	public void update()
	{
		if (state!=0)
		{
			pkgQueue.incoming.add(new SimDataPacket(this,SimKernel.theKernel.currentTime));
		}
		pkgQueue.recvMessages();
		bufferSize+=state+recvQueue; // add new messages to buffer
		usage+=recvQueue; // add recv energy to equation
		recvQueue=0; // reset the recv queues
		
		updateBufferSizeVars();
	}
}