import java.text.DecimalFormat;
import java.util.*;

class TraceEntry
{
	double cumCost;
	double instCost;
	public TraceEntry(double cumCost, double instCost)
	{
		this.cumCost=cumCost;
		this.instCost=instCost;
	}
}

class AttributeSummary
{
	double min;
	double max;
	double median;
	double mean;
	Vector<Double> values;
	AttributeSummary()
	{
		min=Double.MAX_VALUE;
		max=-Double.MAX_VALUE;
		median=0;
		values=new Vector<Double>();
	}
	void addVal(double val)
	{
		values.add(val);
		if (val<min)
			min=val;
		if (val>max)
			max=val;
		Collections.sort(values);
		median=values.get(values.size()/2);
		mean=0;
		for (int i=0;i<values.size();i++)
			mean+=values.get(i);
		mean/=values.size();
	}
}
/**
 * This class holds statistics for multiple runs. 
 * Currently 'average latency' and 'maximum energy use' is used
 *  
 * @author osoysal
 *
 */
class MultiRunStatistics
{
	boolean printTotalEnergy;
	Vector<Statistics> runStats;
	AttributeSummary maxEnergy;
	AttributeSummary averageLatency;
	AttributeSummary totalCost;
	
	public MultiRunStatistics(boolean printTotalEnergy)
	{
		this.printTotalEnergy=printTotalEnergy;
		runStats=new Vector<Statistics>();
		maxEnergy=new AttributeSummary();
		averageLatency=new AttributeSummary();
		totalCost=new AttributeSummary();
	}
	
	void addStatistics(Statistics newStats)
	{
		runStats.add(newStats);
	}
	
	void printCouple(AttributeSummary x, AttributeSummary y)
	{
		System.out.printf("%f %f %f %f %f %f\n",x.median,y.median,x.min,x.max,y.min,y.max);
	}
	
	void printResult()
	{
		for (int i=0;i<runStats.size();i++)
		{
			maxEnergy.addVal(runStats.get(i).maxEnergyUsage);
			averageLatency.addVal(runStats.get(i).avgLatency);
			totalCost.addVal(runStats.get(i).trace.get(runStats.get(i).trace.size()-1).cumCost);
		}
		if (!printTotalEnergy)
			printCouple(maxEnergy,averageLatency);
		else
			System.out.printf("%f,%f,%f,\n",totalCost.mean,totalCost.min,totalCost.max);
	}
}

class Statistics 
{
	SimKernel krn;
	double avgLatency;
	double maxLatency;
	double minLatency;
	
	double maxEnergyUsage;
	double minEnergyUsage;
	double avgEnergyUsage;
	
	Vector<TraceEntry> trace; // cumulative cost of operation (salmon)
	double firstOver5000;
	
	boolean isFilled;
	boolean allBuffersEmpty;
	
	Vector<SimDataPacket> allPackages; 
	
	

	Statistics() 
	{
		krn = SimKernel.theKernel;
		isFilled=false;
		trace=new Vector<TraceEntry>();
		allPackages=new Vector<SimDataPacket>();
		firstOver5000=0;
	}
	
	void reset() // resets the statistics object
	{
		trace.clear();
		isFilled=false;
		allPackages.clear();
		krn.reset();
	}
	
	void stepStatistics() // i is the current statistics step
	{
		// update buffer condition
		allBuffersEmpty=true; 
		for (int j=0;j<krn.nodeTable.length;j++)
		{
			if (krn.nodeTable[j].bufferSize>0)
				allBuffersEmpty=false;
		}
		
		// update lifetime condition
		for (int j=0;j<krn.nodeTable.length;j++)
		{
			SimTreeLeaf n=krn.nodeTable[j];
			if (firstOver5000<0 && n.usage>5000)
				firstOver5000=krn.getSecondsPassed();
		}
		
		// insert trace elements 

		TraceEntry t;
		if (trace.size()==0)
			t=new TraceEntry(krn.getCost(krn.theFish.currentNode),krn.getCost(krn.theFish.currentNode));
		else
		{
			double cost=krn.getCost(krn.theFish.currentNode);
			t=new TraceEntry(trace.get((int)trace.size()-1).cumCost+cost,cost);
		}
		trace.add(t);
	}
	
	void runForStatistics()
	{
		runForStatistics(false);
	}
	
	void runForStatistics(boolean emptyBuffers)
	{
		long tmax = simulationDuration();
		krn.reset();
		trace.clear();
		firstOver5000=-1;
		for (long i = 0; i < tmax || (emptyBuffers && !allBuffersEmpty); i++) 
		{
			krn.stepTime(1); // step simulation
/*
			if (i==tmax-1) // no more data
			{  
				for (int j=0;j<krn.nodeTable.length;j++)
						krn.nodeTable[j].state=0; // reset states of nodes				
			}*/
		}
		// post process package routing information
		double totalDelay=0;
		maxLatency=0;
		for (int i=0;i<allPackages.size();i++)
		{
			totalDelay+=allPackages.get(i).deliveryTime-allPackages.get(i).startTime;
			if (allPackages.get(i).deliveryTime-allPackages.get(i).startTime >maxLatency)
				maxLatency=allPackages.get(i).deliveryTime-allPackages.get(i).startTime ;
		}
		totalDelay/=1000; // to seconds
		maxLatency/=1000; // to seconds
		avgLatency=totalDelay/allPackages.size();

		isFilled=true;
	}
	
	long simulationDuration()
	{
		return (long) Math.ceil((krn.eventList[krn.eventList.length - 1].pctimestamp - krn.initTime) / 1000.0) + 1; // in
		// terms
		// of
		// seconds
	}

	public double getConstantTimeTotalCost() 
	{
		double tmp=Salmon.v; // save it to fix later
		Salmon.v=0.0000000000001;// should be small enough
		runForStatistics();
		Salmon.v=tmp;
		return trace.get(trace.size()-1).cumCost;
	}

	public double getConstantTimeTotalCostMobile() 
	{
		runForStatistics();
		return trace.get(trace.size()-1).cumCost;
	}

	void printXYMatrix() 
	{
		System.out.println("\\xymatrix@=0.05in{");
		for (int i = 0; i < krn.nRows; i++) {
			for (int j = 0; j < krn.RowSize; j++) {
				SimTreeLeaf l = krn.nodeTable[i*krn.RowSize+j];
				System.out.print("{\\bullet}");
				if (l.parent != null) {
					// System.out.println(""+l.parent.r+","+l.parent.c+"
					// "+l.r+","+l.c);
					System.out.print("\\ar[");
					if (Math.abs(l.parent.pos.x - l.pos.x)<0.1)
						System.out.print("0");
					else if ((l.parent.pos.x - l.pos.x)<0)
						System.out.print("-1");
					else 
						System.out.print("1");
					System.out.print(",");
					if (Math.abs(l.parent.pos.y - l.pos.y)<0.1)
						System.out.print("0");
					else if ((l.parent.pos.y - l.pos.y)<0)
						System.out.print("-1");
					else 
						System.out.print("1");
					System.out.print("]");
				}
				if (j < krn.RowSize - 1)
					System.out.print("&");
			}
			System.out.println("\\\\");
		}
		System.out.println("}");
	}

	void printCostTrace() 
	{
		double tmp=Salmon.v;
		Vector<TraceEntry> mobileTrace=new Vector<TraceEntry>();
		Vector<TraceEntry> staticTrace=new Vector<TraceEntry>();
		runForStatistics();
		for (int i=0;i<trace.size();i++)
			mobileTrace.add(trace.get(i));
		Salmon.v=0.0000000001;
		runForStatistics();
		Salmon.v=tmp;
		for (int i=0;i<trace.size();i++)
			staticTrace.add(trace.get(i));
		
		for (int i = 0; i < trace.size(); i++) 
		{
			System.out.println("" + staticTrace.get(i).cumCost+ "," + mobileTrace.get(i).cumCost);
		}
	}

	void printInstantenousCostTrace() 
	{
		double tmp=Salmon.v;
		Vector<TraceEntry> mobileTrace=new Vector<TraceEntry>();
		Vector<TraceEntry> staticTrace=new Vector<TraceEntry>();
		runForStatistics();
		for (int i=0;i<trace.size();i++)
			mobileTrace.add(trace.get(i));
		Salmon.v=0.0000000001;
		runForStatistics();
		Salmon.v=tmp;
		for (int i=0;i<trace.size();i++)
			staticTrace.add(trace.get(i));
		
		for (int i = 0; i < trace.size(); i++) 
		{
			System.out.println("" + staticTrace.get(i).instCost+ "," + mobileTrace.get(i).instCost);
		}
	}

	public void printBurstyTrace() 
	{ // i'm leaving this as is
		// it is not very nice 
		// but i don't care
		long tmax = simulationDuration();
		krn.reset();
		double mobilecost = 0, staticcost = 0;
		SimTreeLeaf opt, oldopt = krn.getOptimalNode();
		for (long i = 0; i < tmax; i++) 
		{
			krn.stepTime(1);
			opt = krn.getOptimalNode();
			mobilecost += krn.getCost(krn.theFish.currentNode);
			staticcost += krn.getCost(krn.treeRoots.get(0).root);
			if (oldopt != opt) 
			{
				mobilecost = 0;
				staticcost = 0;
			}
			oldopt = opt;
			System.out.println("" + staticcost + "," + mobilecost);
		}
	}

	public void printNodeUsage(boolean print)
	{
		runForStatistics(true);
/*		for (int i=0;i<krn.nRows;i++)
		{
			for (int j=0;j<krn.RowSize;j++)
				System.out.print(krn.nodeTable[i*krn.RowSize+j].usage+" ");
			System.out.println();
		}*/
		maxEnergyUsage=0;
		double totalUsage=0;
		for (int i=0;i<krn.nodeTable.length;i++)
		{
			if (maxEnergyUsage<krn.nodeTable[i].usage)
				maxEnergyUsage=krn.nodeTable[i].usage;
			totalUsage+=krn.nodeTable[i].usage;
		}
		double averageUsage=totalUsage/(krn.nRows*krn.RowSize);
		DecimalFormat d=new DecimalFormat("##########.00");
//		d.setMinimumIntegerDigits(6);
		if (print)
		{
			System.out.println("Max Usage:"+d.format(maxEnergyUsage)+" Total Usage:"+d.format(totalUsage)+" MAX/Average:"+d.format(maxEnergyUsage/averageUsage)+" Network Life time:"+d.format(firstOver5000)+"sec");
			System.out.printf("Packages routed: %5d AverageDelay: %5.2f MaxDelay: %5.0f\n",allPackages.size(),avgLatency,maxLatency);
			System.out.printf("Summary: %7.2f %7.2f \n",maxEnergyUsage,avgLatency);
		}
	}

}